/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* rc-package-info.c
 * Copyright (C) 2000, 2001 Ximian, Inc.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include <gnome-xml/xmlmemory.h>
#include <gnome-xml/parser.h>

#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include "rc-world.h"
#include "rc-package-info.h"
#include "rc-debman-general.h"
#include "rc-debug.h"
#include "rc-util.h"
#include "xml-util.h"
#include "rc-package-update.h"
#include "rc-xml.h"

typedef enum _DepType DepType;

enum _DepType {
    DEP_TYPE_REQUIRES,
    DEP_TYPE_PROVIDES,
    DEP_TYPE_CONFLICTS,
    DEP_TYPE_SUGGESTS,
    DEP_TYPE_RECOMMENDS
};

static guint rc_channel_parse_helix (RCChannel *, gchar *);
static guint rc_channel_parse_debian (RCChannel *, gchar *);
static guint rc_channel_parse_redhat (RCChannel *, gchar *);

guint
rc_channel_parse (RCChannel *channel, gchar *tbuf, int compressed_length)
{
    gchar *buf;
    GByteArray *byte_array = NULL;
    guint ret = 0;

    g_assert (tbuf);

    if (compressed_length) {
        if (rc_uncompress_memory (tbuf, compressed_length, &byte_array)) {
            g_warning ("Uncompression failed");
            return 1;
        }

        buf = byte_array->data;
    } else {
        buf = tbuf;
    }

    switch (channel->type) {
    case RC_CHANNEL_TYPE_HELIX:
        ret = rc_channel_parse_helix (channel, buf);
        break;
    case RC_CHANNEL_TYPE_DEBIAN:
        ret = rc_channel_parse_debian (channel, buf);
        break;
    case RC_CHANNEL_TYPE_REDHAT:
        ret = rc_channel_parse_redhat (channel, buf);
        break;
    default:
        g_error ("Trying to parse unknown channel type!");
        break;
    }

    if (byte_array) {
        g_byte_array_free (byte_array, TRUE);
    }

    return (ret);
}

static guint
rc_channel_parse_helix (RCChannel *channel, gchar *buf)
{
    xmlDoc *xml_doc;
    xmlNode *root;
    gboolean err = FALSE;

    g_assert (buf);

    /*
     * For some reason, either because of differences in the RCWorld or
     * because we're using libxml1 instead of libxml2, the SAX parser is
     * much slower than the DOM parser, so use it by default.
     */

    if (getenv ("RC_SAX_XML")) {
        RCPackageSAXContext *ctx;
        RCPackageSList *packages;
        RCWorld *world = rc_get_world();
        guint count;

        ctx = rc_package_sax_context_new(channel);
        rc_package_sax_context_parse_chunk(ctx, buf, strlen(buf));
        packages = rc_package_sax_context_done(ctx);

        count = g_slist_length(packages);

        rc_world_freeze (world);
        g_hash_table_freeze (channel->packages);
        g_hash_table_freeze (channel->dep_table);
        g_hash_table_freeze (channel->dep_name_table);
        
        while (packages) {
            const RCPackageDepSList *prov_iter;
            RCPackage *package;
            
            package = (RCPackage *) packages->data;

            if (rc_world_add_package (world, package)) {
                for (prov_iter = package->provides; prov_iter;
                     prov_iter = prov_iter->next)
                {
                    const RCPackageDep *dep =
                        (RCPackageDep *)(prov_iter->data);
                
                    rc_hash_table_slist_insert_unique (channel->dep_table,
                                                       (gpointer) &dep->spec,
                                                       package, NULL);
                    rc_hash_table_slist_insert_unique (channel->dep_name_table,
                                                       dep->spec.name,
                                                       package, NULL);
                }
            
                g_hash_table_insert (channel->packages,
                                     (gpointer) package->spec.name,
                                     (gpointer) package);
            }
            
            packages = packages->next;
        }
        
        g_hash_table_thaw (channel->packages);
        g_hash_table_thaw (channel->dep_table);
        g_hash_table_thaw (channel->dep_name_table);
        rc_world_thaw (world);
        
        g_slist_free (packages);

        if (count == 0)
            err = TRUE;
    }
    else {
        xml_doc = xmlParseMemory (buf, strlen (buf));
        
        if (!xml_doc) {
            return -1;
        }
        
        root = xmlDocGetRootElement (xml_doc);
        err = rc_xml_node_to_channel (channel, root);
        
        xmlFreeDoc (xml_doc);
    }

    return err;
}

static int
fill_debian_package (RCPackage  *pkg,
                     const char *buf,
                     const char *url_prefix)
{
    const char *ibuf;
    RCPackageUpdate *up = NULL;

    up = rc_package_update_new ();

    ibuf = buf;
    while (1) {
        char *key;
        GString *value = NULL;
        const char *p;
        int ind;

        /* Linebreaks indicate the end of a package block. */
        if (*ibuf == '\0' || *ibuf == '\n') break;

        p = strchr (ibuf, ':');

        /* Something bad happened, we're supposed to have a colon. */
        if (!p) break;

        /* Copy the name of the key and lowercase it */
        key = g_strndup (ibuf, p - ibuf);
        g_strdown (key);
        
        /* Move past the colon and any spaces */
        ibuf = p;
        while (*ibuf && (*ibuf == ':' || *ibuf == ' ')) ibuf++;

        ind = 0;
        while ((p = strchr (ibuf, '\n'))) {
            char *v;

            if (!value)
                value = g_string_new ("");

            v = g_strndup (ibuf, p - ibuf);
            g_string_append (value, v);
            g_free (v);
            ind += p - ibuf;

            ibuf = p;

            /* Move past the newline */
            ibuf++;

            /* Check to see if this is a continuation of the previous line */
            if (*ibuf == ' ') {
                /* It is.  Move past the space */
                ibuf++;

                /*
                 * This is a hack.  Description is special because it's
                 * intended to be multiline and user-visible.  So if we're
                 * dealing with description, add a newline.
                 */

                if (strncmp (key, "description",
                             strlen ("description")) == 0) {
                    g_string_append_c (value, '\n');

                    /*
                     * A period on a line by itself indicates that it
                     * should be a blank line.  A newline will follow the
                     * period, so we'll just skip over it.
                     */
                    if (*ibuf == '.')
                        ibuf++;
                }
            }
            else {
                /* It isn't.  Break out. */
                break;
            }
        }

        if (!strncmp (key, "package", strlen ("package"))) {
            pkg->spec.name = g_strdup (value->str);
        } else if (!strncmp (key, "installed-size",
                             strlen ("installed-size"))) {
            up->installed_size = strtoul (value->str, NULL, 10) * 1024;
        } else if (!strncmp (key, "size", strlen ("size"))) {
            up->package_size = strtoul(value->str, NULL, 10);
        } else if (!strncmp (key, "description", strlen ("description"))) {
            char *newline;

            /*
             * We only want the first line for the summary, and all the
             * other lines for the description.
             */

            newline = strchr (value->str, '\n');
            if (!newline) {
                pkg->summary = g_strdup (value->str);
                pkg->description = g_strconcat (value->str, "\n", NULL);
            }
            else {
                pkg->summary = g_strndup (value->str, newline - value->str);
                pkg->description = g_strconcat (newline + 1, "\n", NULL);
            }
        } else if (!strncmp (key, "version", strlen ("version"))) {
            guint epoch;
            rc_debman_parse_version (value->str,
                                     &epoch, &pkg->spec.version,
                                     &pkg->spec.release);
            pkg->spec.epoch = epoch;
        } else if (!strncmp (key, "section", strlen ("section"))) {
            pkg->section = rc_debman_section_to_package_section (value->str);
        } else if (!strncmp (key, "depends", strlen ("depends"))) {
            pkg->requires = g_slist_concat (
                pkg->requires,
                rc_debman_fill_depends (value->str));
        } else if (!strncmp (key, "recommends", strlen ("recommends"))) {
            pkg->recommends = g_slist_concat (
                pkg->recommends,
                rc_debman_fill_depends (value->str));
        } else if (!strncmp (key, "suggests", strlen ("suggests"))) {
            pkg->suggests = g_slist_concat (
                pkg->suggests,
                rc_debman_fill_depends (value->str));
        } else if (!strncmp (key, "pre-depends", strlen ("pre-depends"))) {
            RCPackageDepSList *tmp, *iter;

            tmp = rc_debman_fill_depends (value->str);

            for (iter = tmp; iter; iter = iter->next) {
                RCPackageDep *dep = (RCPackageDep *) iter->data;

                dep->pre = TRUE;
            }

            pkg->requires = g_slist_concat (pkg->requires, tmp);
        } else if (!strncmp (key, "conflicts", strlen ("conflicts"))) {
            pkg->conflicts = g_slist_concat (
                pkg->conflicts,
                rc_debman_fill_depends (value->str));
        } else if (!strncmp (key, "provides", strlen ("provides"))) {
            pkg->provides = g_slist_concat (
                pkg->provides,
                rc_debman_fill_depends (value->str));
        } else if (!strncmp (key, "filename", strlen ("filename"))) {
            /* Build a new update with just this version */
            if (url_prefix) {
                up->package_url = g_strconcat (url_prefix, "/",
                                               value->str,
                                               NULL);
            } else {
                up->package_url = g_strdup (value->str);
            }
        } else if (!strncmp (key, "md5sum", strlen ("md5sum"))) {
            up->md5sum = g_strdup (value->str);
        } else if (!strncmp (key, "architecture", strlen ("architecture"))) {
            pkg->arch = rc_arch_from_string (value->str);
        }

        g_free (key);
        g_string_free (value, TRUE);
    }

    up->importance = RC_IMPORTANCE_SUGGESTED;
    up->description = g_strdup ("Upstream Debian release");
    rc_package_spec_copy (&up->spec, &pkg->spec);
    rc_package_add_update (pkg, up);

    /* Make sure to provide myself, for the dep code! */
    pkg->provides =
        g_slist_append (pkg->provides,
                        rc_package_dep_new_from_spec (&pkg->spec,
                                                      RC_RELATION_EQUAL));

    /* returns the number of characters we processed */
    return ibuf - buf;
}

static guint
rc_channel_parse_debian (RCChannel *rcc, char *buf)
{
    /* Note that buf is NOT ours -- we can't munge it, and strdup'ing the
     * whole thing would be ufly. So, we let packages_query_helper handle
     */
    RCWorld *world = rc_get_world ();
    const char *iter;

    g_hash_table_freeze (rcc->packages);
    g_hash_table_freeze (rcc->dep_table);
    g_hash_table_freeze (rcc->dep_name_table);

    /* Keep looking for a "Packages: " */
    iter = buf;
    while ((iter = strstr (iter, "Package: ")) != NULL) {
        RCPackage *p;
        int off;
        GSList *ps;

        p = rc_package_new ();

        /* All debian packages "have" epochs */
        p->spec.has_epoch = TRUE;

        off = fill_debian_package (p, iter, rcc->file_path);

        p->channel = rcc;

        ps = p->provides;
        while (ps) {
            RCPackageDep *dep = (RCPackageDep *) ps->data;
            rc_hash_table_slist_insert_unique (rcc->dep_table, &dep->spec, p, NULL);
            rc_hash_table_slist_insert_unique (rcc->dep_name_table, dep->spec.name, p, NULL);
            ps = ps->next;
        }

        if (rc_world_add_package (world, p)) {
            g_hash_table_insert (rcc->packages, p->spec.name, p);
            rc_hash_table_slist_insert_unique (rcc->dep_table,
                                               &p->spec, p, NULL);
            rc_hash_table_slist_insert_unique (rcc->dep_name_table,
                                               p->spec.name, p, NULL);
        }

        iter += off;
    }

    g_hash_table_thaw (rcc->packages);
    g_hash_table_thaw (rcc->dep_table);
    g_hash_table_thaw (rcc->dep_name_table);

    return 0;
}

static guint
rc_channel_parse_redhat (RCChannel *rcc, char *buf)
{
    g_error ("No redhat channel support!");
    return 0;
}

/* Helper functions */

RCPackage *
pkginfo_find_package_with_constraint (RCChannel *ch,
                                      gchar *name,
                                      RCPackageDep *constraint,
                                      gboolean is_and)

{
    RCPackage *rp, *candidate = NULL;
    RCPackageUpdate *upi;

    g_return_val_if_fail (ch != NULL, NULL);
    
#if 0
    rc_debug (RC_DEBUG_LEVEL_DEBUG, "pfpwc: name: %s constr: ", name);
    rc_package_dep_dump (constraint);
#endif

    rp = g_hash_table_lookup (ch->packages, name);

    if (rp) {
#if DEBUG > 10
        rc_debug (RC_DEBUG_LEVEL_DEBUG, "   found %s in hash...\n", rp->spec.name);
#endif
        if (constraint) {
            RCPackageDep *dep;

            g_assert(rp->history);
            g_assert(rp->history->data);
            
            upi = (RCPackageUpdate *) rp->history->data;

            dep = rc_package_dep_new_from_spec (&(upi->spec),
                                                RC_RELATION_EQUAL);

            if (rc_package_dep_verify_relation
                (constraint, dep))
                {
                    candidate = rp;
                }

            rc_package_dep_free (dep);
        }
        else {
            candidate = rp;
        }
        if (candidate)
            return candidate;
#if DEBUG > 10
        if (constraint)
            rc_debug (RC_DEBUG_LEVEL_DEBUG, "   failed constraint test.\n");
        else
            rc_debug (RC_DEBUG_LEVEL_DEBUG, "   didn't exist!\n");
#endif
    }

    return NULL;
}

RCPackage *
pkginfo_find_package (RCChannel *chs,
                      gchar *name)
{
    return pkginfo_find_package_with_constraint (chs, name, NULL, FALSE);
}
