/* -*- Mode: C; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/*
 * gui-transfer.c: Wrapper around GnomeTransfer that does blocking transfers
 * and handles caching
 *
 * Copyright (C) 2000, 2001 Ximian, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */

#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

#include <gnome.h>
#include <libsoup/soup.h>
#include "gui.h"
#include "gui-transfer.h"
#include "gui-priority.h"
#include "gui-prefs.h"
#include "gui-cache.h"
#include "gui-util.h"

static GtkObjectClass *gui_transfer_parent;

enum SIGNALS {
    FILE_DATA,
    FILE_DONE,
    LAST_SIGNAL
};

#define BLOCK_SIZE 8192

#define HTTP_RESPONSE_SUCCESSFUL(x)   (x >= 200 && x < 300)
#define HTTP_RESPONSE_REDIRECT(x)     (x >= 300 && x < 400)
#define HTTP_RESPONSE_CLIENT_ERROR(x) (x >= 400 && x < 400)
#define HTTP_RESPONSE_SERVER_ERROR(x) (x >= 500 && x < 600)

#define HTTP_RESPONSE_AUTH_FAILURE(x) (x == 401 || x == 407)
#define HTTP_RESPONSE_NOT_MODIFIED(x) (x == 304)

/* GSList of GUITransfers */
GSList *current_transfers = NULL;

static guint signals[LAST_SIGNAL] = { 0 };

static void
gui_transfer_destroy(GtkObject *obj)
{
    GUITransfer *t = GUI_TRANSFER(obj);
    
    g_free(t->filename);
    g_free(t->url);
    if (t->timer)
        g_timer_destroy(t->timer);
    if (t->error_string)
        g_free(t->error_string);

    if (GTK_OBJECT_CLASS(gui_transfer_parent)->destroy)
        (* GTK_OBJECT_CLASS(gui_transfer_parent)->destroy)(obj);
} /* gui_transfer_destroy */

static void
gui_transfer_class_init(GUITransferClass *klass)
{
    GtkObjectClass *object_class = (GtkObjectClass *) klass;

    object_class->destroy = gui_transfer_destroy;

    gui_transfer_parent = gtk_type_class(gtk_object_get_type());

    signals[FILE_DATA] = 
        gtk_signal_new(
            "file_data",
            GTK_RUN_LAST,
            object_class->type,
            GTK_SIGNAL_OFFSET(GUITransferClass, file_data),
            gtk_marshal_NONE__POINTER_INT,
            GTK_TYPE_NONE, 2,
            GTK_TYPE_POINTER,
            GTK_TYPE_INT);

    signals[FILE_DONE] =
        gtk_signal_new(
            "file_done",
            GTK_RUN_LAST,
            object_class->type,
            GTK_SIGNAL_OFFSET(GUITransferClass, file_done),
            gtk_marshal_NONE__NONE,
            GTK_TYPE_NONE, 0);

    gtk_object_class_add_signals(object_class, signals, LAST_SIGNAL);
} /* gui_transfer_class_init */

static void
gui_transfer_init(GUITransfer *t)
{
    char *purl;

    purl = gui_prefs_get_proxy_url();
    gui_transfer_set_proxy_url(t, purl);
    g_free(purl);

    t->error = GUI_TRANSFER_ERROR_NONE;
    t->cache = NULL;
} /* gui_transfer_init */

guint
gui_transfer_get_type(void)
{
    static guint type = 0;

    if (!type) {
        GtkTypeInfo type_info = {
            "GUITransfer",
            sizeof(GUITransfer),
            sizeof(GUITransferClass),
            (GtkClassInitFunc) gui_transfer_class_init,
            (GtkObjectInitFunc) gui_transfer_init,
            (GtkArgSetFunc) NULL,
            (GtkArgGetFunc) NULL
        };

        type = gtk_type_unique(gtk_object_get_type(), &type_info);
    }

    return type;
} /* gui_transfer_get_type */

GUITransfer *
gui_transfer_new(GUITransferFlags flags, GUICache *cache)
{
    GUITransfer *t;

    t = GUI_TRANSFER(gtk_type_new(gui_transfer_get_type()));

    t->flags = flags;
    t->cache = cache;

    return t;
} /* gui_transfer_new */

void
gui_transfer_set_flags(GUITransfer *t, GUITransferFlags flags)
{
    g_return_if_fail(t);
    g_return_if_fail(GUI_IS_TRANSFER(t));

    t->flags = flags;
} /* gui_transfer_set_flags */

void
gui_transfer_set_proxy_url(GUITransfer *t, const char *url)
{
    SoupContext *context;

    if (url) {
        context = soup_context_get(url);
        soup_set_proxy(context);
    }
    else
        soup_set_proxy(NULL);
} /* gui_transfer_set_proxy_info */

void
gui_transfer_set_progress_bar(GUITransfer *t, GtkProgress *progress_bar)
{
    g_return_if_fail(t);
    g_return_if_fail(GUI_IS_TRANSFER(t));
    g_return_if_fail(progress_bar);
    g_return_if_fail(GTK_IS_PROGRESS(progress_bar));

    t->progress_bar = GTK_WIDGET(progress_bar);
} /* gui_transfer_set_progress_bar */

void
gui_transfer_set_progress_label(GUITransfer *t, GtkLabel *label)
{
    g_return_if_fail(t);
    g_return_if_fail(GUI_IS_TRANSFER(t));
    g_return_if_fail(label);
    g_return_if_fail(GTK_IS_LABEL(label));

    t->label = GTK_WIDGET(label);
} /* gui_transfer_set_progress_label */

GSList *
gui_transfer_get_current_transfers(void)
{
    return current_transfers;
} /* gui_transfer_get_current_transfers */

GUITransferError
gui_transfer_get_error(GUITransfer *t)
{
    return t->error;
} /* gui_transfer_get_error */

const char *
gui_transfer_get_error_string(GUITransfer *t)
{
    g_return_val_if_fail(t, NULL);
    g_return_val_if_fail(t->error, NULL);

    return t->error_string;
} /* gui_transfer_get_error_string */

static const char *
error_to_string(GUITransfer *t)
{
    char *err;

    switch (t->error) {
    case GUI_TRANSFER_ERROR_NONE:
        err = "(no error)";
        break;
    case GUI_TRANSFER_ERROR_CANCELLED:
        err = "Cancelled";
        break;
    case GUI_TRANSFER_ERROR_CANT_CONNECT:
        err = "Can't connect";
        break;
    case GUI_TRANSFER_ERROR_FILE_NOT_FOUND:
        err = "File not found";
        break;
    case GUI_TRANSFER_ERROR_IO:
        err = "IO error";
        break;
    case GUI_TRANSFER_ERROR_CANT_AUTHENTICATE:
        err = "Unable to authenticate";
        break;
    case GUI_TRANSFER_ERROR_INVALID_URI:
        err = "Invalid URI";
        break;
    default:
        err = NULL;
        g_assert_not_reached();
        break;
    }
      
    return err;
} /* error_to_string */

static void
gui_transfer_set_error(GUITransfer *t, GUITransferError err,
                       const char *err_string)
{
    const char *e;

    t->error = err;
    g_free(t->error_string);

    e = error_to_string(t);

    if (err_string)
        t->error_string = g_strdup_printf("%s - %s", e, err_string);
    else
        t->error_string = g_strdup(e);
} /* gui_transfer_set_error */

static char *
format_size(guint size)
{
    char *output;

    if (size > 1024*1024) {
        output = g_strdup_printf(
            _("%.2f MB"), (float) size / ((float) 1024*1024));
    }
    else if (size > 1024)
        output = g_strdup_printf(_("%d KB"), size / 1024);
    else
        output = g_strdup_printf(_("%d bytes"), size);

    return output;
} /* format_size */

static void
gui_transfer_emit_data(GUITransfer *t, char *buf, int size)
{
    t->trans_size += size;

    if (!(t->flags & GUI_TRANSFER_FLAGS_FLUSH_MEMORY)) {
        if (!t->data)
            t->data = g_byte_array_new();
        
        t->data = g_byte_array_append(t->data, buf, size);
    }

    if (t->progress_bar) {
        double elapsed_time;
        char *et_str = NULL;
        char *s1, *s2;
        char *pstr;

        elapsed_time = g_timer_elapsed(t->timer, NULL);
        /* Give the timer a second before we start estimating throughput */
        if (elapsed_time > 1.0)
            et_str = format_size((guint) t->trans_size/elapsed_time);

        if (t->file_size > 0) {
            s1 = format_size(t->trans_size + t->offset);
            s2 = format_size(t->file_size);
            if (et_str) {
                pstr = g_strdup_printf(
                    _("%%p%%%% (%s of %s) - %s/s"), s1, s2, et_str);
                g_free(et_str);
            }
            else
                pstr = g_strdup_printf(_("%%p%%%% (%s of %s)"), s1, s2);

            g_free(s1);
            g_free(s2);
        }
        else {
            s1 = format_size(t->trans_size + t->offset);
            if (et_str) {
                pstr = g_strdup_printf(_("%s - %s/s"), s1, et_str);
                g_free(et_str);
            }
            else
                pstr = g_strdup_printf("%s", s1);

            g_free(s1);
        }

        gtk_progress_set_format_string(GTK_PROGRESS(t->progress_bar), pstr);
        g_free(pstr);
        gtk_progress_set_value(
            GTK_PROGRESS(t->progress_bar),
            (gfloat) (t->trans_size + t->offset));
    }

    gtk_signal_emit(GTK_OBJECT(t), signals[FILE_DATA], buf, size);
} /* gui_transfer_emit_data */

static void
gui_transfer_emit_done(GUITransfer *t)
{
    current_transfers = g_slist_remove(current_transfers, t);

    if (!gui_transfer_get_error(t)) {
        g_timer_stop(t->timer);
    }
    gtk_signal_emit(GTK_OBJECT(t), signals[FILE_DONE]);

    if (t->flags & GUI_TRANSFER_FLAGS_BLOCK) {
        gtk_main_quit();
    }
    else {
        /* If this transfer is non-blocking, we want to free up the memory
           in the buffer. */
        if (t->data)
            g_byte_array_free(t->data, TRUE);
        t->data = NULL;
    }
} /* gui_transfer_emit_done */

static void
gui_transfer_file_info(GUITransfer *t)
{
    if (!t->timer)
        t->timer = g_timer_new();
    else
        g_timer_reset(t->timer);

    g_timer_start(t->timer);

    /* Set the text of the label, if there is one */
    if (t->label)
        gtk_label_set_text(GTK_LABEL(t->label), _("Downloading data..."));

    /* Set the value and limits of the progress bar, if there is one */
    if (t->progress_bar) {
        if (t->file_size > 0) {
            gtk_progress_set_activity_mode(
                GTK_PROGRESS(t->progress_bar), FALSE);
            gtk_progress_configure(
                GTK_PROGRESS(t->progress_bar), 0.0, 0.0, t->file_size);
        }
        else {
            gtk_progress_set_activity_mode(
                GTK_PROGRESS(t->progress_bar), TRUE);
            gtk_progress_configure(
                GTK_PROGRESS(t->progress_bar), 0.0, 0.0, 100.0);
        }
    }
} /* gui_transfer_file_info */

#if 0
static int
parse_month(char *month)
{
    char *months[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug",
                       "Sep", "Oct", "Nov", "Dec" };
    int i;

    for (i = 0; i < 12; i++) {
        if (g_strncasecmp(month, months[i], 3) == 0)
            return i;
    }

    g_assert_not_reached();
    return 0;
} /* parse_month */

static time_t
parse_http_time(const char *header)
{
    struct tm *tm;
    time_t tt;
    char **split;
    char **time;
    char **date;

    g_return_val_if_fail(header, -1);

    /* According to RFC 2616, all HTTP 1.1 compliant clients must be able to
       parse dates in any one of the following three formats:

       Sun, 06 Nov 1994 08:49:37 GMT  ; RFC 822, updated by RFC 1123
       Sunday, 06-Nov-94 08:49:37 GMT ; RFC 850, obsoleted by RFC 1036 
       Sun Nov  6 08:49:37 1994       ; ANSI C's asctime() format

       The key, as you can see, is the 4th character, which will tell us
       which format we are in. */

    tm = g_new0(struct tm, 1);

    if (header[3] == ',') {
        /* Sun, 06 Nov 1994 08:49:37 GMT */

        split = g_strsplit(header, " ", 0);
        tm->tm_mday = atoi(split[1]);
        tm->tm_mon = parse_month(split[2]);
        tm->tm_year = atoi(split[3]) - 1900;
        time = g_strsplit(split[4], ":", 0);
        tm->tm_hour = atoi(time[0]);
        tm->tm_min = atoi(time[1]);
        tm->tm_sec = atoi(time[2]);
        g_strfreev(time);
        g_strfreev(split);
    }
    else if (header[3] == ' ') {
        /* Sun Nov  6 08:49:37 1994 */
        int i;

        split = g_strsplit(header, " ", 0);
        tm->tm_mon = parse_month(split[1]);
        i = 2;
        if (!split[i][0])
            i++;
        tm->tm_mday = atoi(split[i++]);
        time = g_strsplit(split[i++], ":", 0);
        tm->tm_hour = atoi(time[0]);
        tm->tm_min = atoi(time[1]);
        tm->tm_sec = atoi(time[2]);
        g_strfreev(time);
        tm->tm_year = atoi(split[i]);
        g_strfreev(split);
    }
    else {
        /* Sunday, 06-Nov-94 08:49:37 GMT */
        int y;

        split = g_strsplit(header, " ", 0);
        date = g_strsplit(split[1], "-", 0);
        tm->tm_mday = atoi(date[0]);
        tm->tm_mon = parse_month(date[1]);
        /* Two digit dates. Wonderful. Lets assume that anything before the 
           epoch is a 21st century date and afterward is 20th. */
        y = atoi(date[2]);
        if (y >= 70)
            tm->tm_year = y;
        else
            tm->tm_year = 100 + y;
        g_strfreev(date);
        time = g_strsplit(split[2], ":", 0);
        tm->tm_hour = atoi(time[0]);
        tm->tm_min = atoi(time[1]);
        tm->tm_sec = atoi(time[2]);
        g_strfreev(time);
        g_strfreev(split);
    }

    tm->tm_isdst = -1;
        
    tt = mktime(tm);
    g_free(tm);
    
    return tt;
} /* parse_http_time */
#endif

/* Check the following conditions:
 *    (1) That this transfer type is cacheable.  http is, but file isn't.
 *    (2) That this transfer is set to flush memory after each chunk.  If it
 *        is, we have to put the data *somewhere*, so it goes to disk even
 *        if caching is turned off.
 *    (3) That caching is enabled.
 */
static gboolean
is_cacheable(GUITransfer *t)
{
    if (!t->cacheable)
        return FALSE;

    if (t->flags & GUI_TRANSFER_FLAGS_FLUSH_MEMORY)
        return TRUE;

    if (gui_prefs_get_cache_enabled() ||
        t->flags & GUI_TRANSFER_FLAGS_FORCE_CACHE)
        return TRUE;
    else
        return FALSE;
} /* is_cacheable */

static gboolean
file_read_data(GIOChannel *iochannel,
               GIOCondition condition,
               gpointer data)
{
    GUITransfer *transfer = data;
    GIOError err;
    char buf[BLOCK_SIZE];
    int bytes;
    int watch;

    watch = GPOINTER_TO_INT(transfer->proto_data);

    if (!(condition & G_IO_IN))
        goto ERROR;

    err = g_io_channel_read(iochannel, buf, BLOCK_SIZE, &bytes);

    if (bytes)
        gui_transfer_emit_data(transfer, buf, bytes);

    /* G_IO_ERROR_AGAIN means that there is more data to read off the channel */
    if (err == G_IO_ERROR_AGAIN)
        return TRUE;

    if (err != G_IO_ERROR_NONE)
        goto ERROR;

    if (bytes > 0) {
        /* More data to read. Whee. */
        return TRUE;
    }
    else {
        /* No more bytes to read and no error condition; the file is done */
        g_io_channel_close(iochannel);
        gui_transfer_emit_done(transfer);

        return FALSE;
    }

ERROR:
    g_source_remove(watch);

    gui_transfer_set_error(transfer, GUI_TRANSFER_ERROR_IO, NULL);
    gui_transfer_emit_done(transfer);

    return FALSE;
} /* file_read_data */

static void
file_pause(GUITransfer *t)
{
    int watch = GPOINTER_TO_INT(t->proto_data);

    rc_debug(RC_DEBUG_LEVEL_INFO, "PAUSE! Trans size: %d\n", t->trans_size);

    g_source_remove(watch);

    t->offset = t->trans_size + t->offset;
    t->trans_size = 0;
} /* file_pause */

static void
file_abort(GUITransfer *t)
{
    int watch = GPOINTER_TO_INT(t->proto_data);

    g_source_remove(watch);

    gui_transfer_set_error(t, GUI_TRANSFER_ERROR_CANCELLED, NULL);
    gui_transfer_emit_done(t);
} /* file_abort */
        
static int
file_open(GUITransfer *t, int offset)
{
    /* Discard the "file://" portion of the URL */
    const char *filename = t->url + 7;
    int fd;
    struct stat fdstat;
    GIOChannel *iochannel;
    int watch;

    if (stat(filename, &fdstat)) {
        gui_transfer_set_error(t, GUI_TRANSFER_ERROR_FILE_NOT_FOUND, NULL);
        return -1;
    }

    fd = open(filename, O_RDONLY);
    if (fd < 0) {
        gui_transfer_set_error(
            t, GUI_TRANSFER_ERROR_IO, _("Couldn't open file for writing"));
        return -1;
    }

    t->file_size = fdstat.st_size;

    gui_transfer_file_info(t);

    iochannel = g_io_channel_unix_new(fd);

    if (t->offset) {
        GIOError err;

        err = g_io_channel_seek(iochannel, t->offset, G_SEEK_SET);

        if (err != G_IO_ERROR_NONE)
            g_warning("Seek error");
    }

    watch = g_io_add_watch(
        iochannel, G_IO_IN | G_IO_ERR | G_IO_HUP | G_IO_NVAL,
        file_read_data, t);
    t->proto_data = GINT_TO_POINTER(watch);

    g_io_channel_unref(iochannel);

    return 0;
} /* file_open */

static void
http_done(SoupMessage *message, SoupErrorCode err, gpointer user_data)
{
    GUITransfer *transfer = user_data;
    GUITransferError gte;
    char *gte_msg = NULL;

    rc_debug(RC_DEBUG_LEVEL_INFO, "[%p]: http_done called\n", message);
    rc_debug(RC_DEBUG_LEVEL_INFO,
             "[%p]: Soup Error code: %d\n", message, err);

    /* The transfer was paused... */
    if (err == SOUP_ERROR_CANCELLED && transfer->paused)
        return;

    /* If the transfer is cached, read it from there now. */
    if (transfer->cached) {
        char *filename;

        filename = gui_cache_get_cache_filename(
            transfer->cache, transfer->filename);
        g_free(transfer->url);
        transfer->url = g_strdup_printf("file://%s", filename);
        g_free(filename);
        
        rc_debug(
            RC_DEBUG_LEVEL_INFO, "[%p]: Reading from the cache: %s\n",
            message, transfer->url);
        
        file_open(transfer, 0);

        rc_debug(RC_DEBUG_LEVEL_INFO, "\n");

        return;
    }

    if (message->response_code && message->response_phrase) {
        gte_msg = g_strdup_printf(
            "%d %s", message->response_code, message->response_phrase);
    }

    /* Mapping of Soup errors to GUITransfer errors. They're very similar... */
    if (err == SOUP_ERROR_CANCELLED)
        gte = GUI_TRANSFER_ERROR_CANCELLED;
    else if (message->response_code == 404)
        gte = GUI_TRANSFER_ERROR_FILE_NOT_FOUND;
    else if (err == SOUP_ERROR_IO)
        gte = GUI_TRANSFER_ERROR_IO;
    else if (err == SOUP_ERROR_MALFORMED_HEADER) {
        gte = GUI_TRANSFER_ERROR_IO;
        gte_msg = g_strdup("Malformed HTTP header");
    }
    else if (err == SOUP_ERROR_CANT_AUTHENTICATE ||
             HTTP_RESPONSE_AUTH_FAILURE(message->response_code))
        gte = GUI_TRANSFER_ERROR_CANT_AUTHENTICATE;
    else if (err == SOUP_ERROR_CANT_CONNECT) {
        rc_debug(RC_DEBUG_LEVEL_INFO,
                 "[%p]: SOUP_ERROR_CANT_CONNECT\n", message);
        gte = GUI_TRANSFER_ERROR_CANT_CONNECT;
        /* CANT_CONNECT && !gte_msg signify a DNS failure or a refused conn */
        if (!gte_msg)
            gte_msg = g_strdup("DNS failure or connection refused");
    }
    else if (err == SOUP_ERROR_HANDLER) {
        rc_debug(RC_DEBUG_LEVEL_INFO,
                 "[%p]: SOUP_ERROR_HANDLER\n", message);
        gte = GUI_TRANSFER_ERROR_CANT_CONNECT;
    }
    else if (err == SOUP_ERROR_NONE && !transfer->data &&
             !(transfer->flags & GUI_TRANSFER_FLAGS_FLUSH_MEMORY)) {
        /* This is a workaround what seems to be a bug in Soup, but I am having
         * a tough time tracking it down in the Soup code...  If we make a
         * regular HTTP request to an HTTPS port (ie, the remote end drops the
         * connection immediately following a request, Soup requeues it.
         * What ultimately happens, though, is that we get a call only to
         * this function, with no error but also no data.  I should think that
         * it would return an IO error or CANT_CONNECT, but it doesn't.
         *
         * So anyway, if transfer->data is NULL but err is SOUP_ERROR_NONE,
         * then set the gte to GUI_TRANSFER_ERROR_CANT_CONNECT
         */
        
        rc_debug(RC_DEBUG_LEVEL_INFO,
                 "[%p]: SOUP_ERROR_NONE, !transfer->data, !FLUSH_MEM\n",
                 message);
        gte = GUI_TRANSFER_ERROR_CANT_CONNECT;
        gte_msg = g_strdup("Connection prematurely dropped");
    }
    else
        gte = GUI_TRANSFER_ERROR_NONE;

    if (is_cacheable(transfer) &&
        gui_cache_is_active(transfer->cache, transfer->filename)) {
        gui_cache_close(transfer->cache, transfer->filename);
    }

    rc_debug(RC_DEBUG_LEVEL_INFO, "\n\n");

    gui_transfer_set_error(transfer, gte, gte_msg);
    gui_transfer_emit_done(transfer);
    g_free(gte_msg);
} /* http_done */

static SoupErrorCode
http_content_length(SoupMessage *message, gpointer data)
{
    GUITransfer *t = data;
    const char *cl;

    /* Content-Range is present and has already been set. Content-Length is
       only a partial size; the full size is in Content-Range */
    if (t->content_range_set)
        return SOUP_ERROR_NONE;

    cl = soup_message_get_response_header(message, "Content-Length");
    t->file_size = atoi(cl);

    rc_debug(
        RC_DEBUG_LEVEL_INFO, "[%p]: Got Content-Length: %s\n",
        message, cl);

    return SOUP_ERROR_NONE;
} /* http_content_length */

static SoupErrorCode
http_content_range(SoupMessage *message, gpointer data)
{
    GUITransfer *t = data;
    const char *cr;
    int total;

    cr = soup_message_get_response_header(message, "Content-Range");
    sscanf(cr, "bytes %*d-%*d/%d", &total);
    t->file_size = total;
    t->content_range_set = TRUE;

    rc_debug(
        RC_DEBUG_LEVEL_INFO, "[%p]: Got Content-Range: %s\n",
        message, cr);

    return SOUP_ERROR_NONE;
} /* http_content_range */

static SoupErrorCode
http_response_not_modified(SoupMessage *message, gpointer data)
{
    GUITransfer *t = data;

    rc_debug(
        RC_DEBUG_LEVEL_INFO, "[%p]: http_response_not_modified called\n",
        message);

    t->cached = TRUE;

    return SOUP_ERROR_NONE;
} /* http_response_not_modified */

static SoupErrorCode
http_response_ok(SoupMessage *message, gpointer data)
{
    GUITransfer *t = data;

    rc_debug(
        RC_DEBUG_LEVEL_INFO, "[%p]: http_response_ok called\n",
        message);

    /* If we asked for a range of bytes and didn't get it... */
    t->partial = FALSE;

    return SOUP_ERROR_NONE;
} /* http_response_ok */

static SoupErrorCode
http_response_partial_content(SoupMessage *message, gpointer data)
{
    GUITransfer *t = data;

    rc_debug(
        RC_DEBUG_LEVEL_INFO, "[%p]: http_partial_content called\n",
        message);

    /* We're getting a partial content body */
    t->partial = TRUE;

    return SOUP_ERROR_NONE;
} /* http_response_partial_content */

static SoupErrorCode
http_info(SoupMessage *message, gpointer data)
{
    GUITransfer *t = data;

    /* If there is an authentication problem, let's just drop it for now.
       Soup has internal handlers that will fix this on its own. */
    if (HTTP_RESPONSE_AUTH_FAILURE(message->response_code))
        return SOUP_ERROR_NONE;

    rc_debug(RC_DEBUG_LEVEL_INFO, "[%p]: http_info called\n", message);
    rc_dump_transfer_info(message);

    /* If we got a 304 Not Modified from the server, we're just going
       to return quietly here, since it'll be pulled from the cache in
       http_done. */
    if (HTTP_RESPONSE_NOT_MODIFIED(message->response_code))
        return SOUP_ERROR_NONE;

    /* We only want to do this on success... */
    /* Set an error code of SOUP_ERROR_NONE so that the read_data function
       is called and we can get the actual error message from the server. */
    if (!HTTP_RESPONSE_SUCCESSFUL(message->response_code)) {
        rc_debug(RC_DEBUG_LEVEL_INFO,
                 "[%p]: http_info - unsuccessful response code: "
                 "%d\n", message, message->response_code);
        return SOUP_ERROR_NONE;
    }

    if (is_cacheable(t))
        gui_cache_open(t->cache, t->filename, t->offset ? TRUE : FALSE);

    gui_transfer_file_info(t);

    return SOUP_ERROR_NONE;
} /* http_info */

static SoupErrorCode
http_read_data(SoupMessage *message, gpointer data)
{
    GUITransfer *t = data;

    /* Debugging output */
    if (!HTTP_RESPONSE_SUCCESSFUL(message->response_code) &&
        message->response.length) {
        char *data = g_strndup(
            message->response.body, message->response.length);
        
        rc_debug(RC_DEBUG_LEVEL_DEBUG,
                 "[%p]: Server data:\n---\n%s\n---\n", message, data);
        g_free(data);
    }

    /* If there is an authentication problem, let's just drop it for now.
       Soup has internal handlers that will fix this on its own. */
    if (HTTP_RESPONSE_AUTH_FAILURE(message->response_code))
        return SOUP_ERROR_NONE;
    
    if (!HTTP_RESPONSE_SUCCESSFUL(message->response_code)) {
        rc_debug(RC_DEBUG_LEVEL_INFO, 
                 "[%p]: http_read_data - unsuccessful response code: "
                 "%d, setting HANDLER\n", message, message->response_code);
        
        return SOUP_ERROR_HANDLER;
    }

    if (is_cacheable(t)) {
        gui_cache_append(
            t->cache, t->filename,
            message->response.body, message->response.length);
    }

    rc_debug(RC_DEBUG_LEVEL_INFO, "[%p]: Appending %d bytes\n",
             message, message->response.length);

    gui_transfer_emit_data(
        t, message->response.body, message->response.length);

    return SOUP_ERROR_NONE;
} /* http_read_data */

static void
http_pause(GUITransfer *t)
{
    SoupMessage *message = t->proto_data;

    rc_debug(RC_DEBUG_LEVEL_INFO, "PAUSE! Trans size: %d\n", t->trans_size);

    soup_message_cancel(message);

    t->proto_data = NULL;

    t->offset = t->trans_size + t->offset;
    t->trans_size = 0;
} /* http_pause */

static void
http_abort(GUITransfer *t)
{
    SoupMessage *message = t->proto_data;

    if (message)
        soup_message_cancel(message);

    if (is_cacheable(t) && gui_cache_is_active(t->cache, t->filename))
        gui_cache_invalidate(t->cache, t->filename);

    t->proto_data = NULL;

    gui_transfer_set_error(t, GUI_TRANSFER_ERROR_CANCELLED, NULL);
    gui_transfer_emit_done(t);
} /* http_abort */

static int
http_open(GUITransfer *t, int offset)
{
    SoupContext *context;
    SoupMessage *message;

    if (appinfo->priority) {
        SoupUri *uri = rc_gui_priority_authorized_url(t->url);

        if (!uri) {
            gui_transfer_set_error(t, GUI_TRANSFER_ERROR_INVALID_URI, t->url);
            return -1;
        }

        context = soup_context_from_uri(uri);
        soup_uri_free(uri);
    }
    else
        context = soup_context_get(t->url);

    /* No context?  Probably a bad URL. */
    if (!context) {
        gui_transfer_set_error(t, GUI_TRANSFER_ERROR_INVALID_URI, t->url);
        return -1;
    }

    message = soup_message_new(context, NULL);

    t->proto_data = message;

    soup_message_set_method(message, SOUP_METHOD_GET);

    if (gui_prefs_get_http10_enabled())
        soup_message_set_http_version(message, SOUP_HTTP_1_0);

    if (!offset && t->flags & GUI_TRANSFER_FLAGS_RESUME_PARTIAL) {
        char *cache_fn;
        char *cache_tmp;

        cache_fn = gui_cache_get_cache_filename(t->cache, t->filename);
        cache_tmp = g_strdup_printf("%s.tmp", cache_fn);

        rc_debug(
            RC_DEBUG_LEVEL_INFO,
            "[%p]: Trying to find a file to resume... ",
            message);

        if (g_file_exists(cache_tmp)) {
            struct stat s;

            stat(cache_tmp, &s);

            t->offset = offset = s.st_size;

            rc_debug(
                RC_DEBUG_LEVEL_INFO, "Found it! File %s, size %d\n",
                cache_tmp, offset);
        }
        else {
            rc_debug(RC_DEBUG_LEVEL_INFO, "Not found\n");
        }

        g_free(cache_fn);
        g_free(cache_tmp);
    }

    if (offset) {
        char *range = g_strdup_printf("bytes=%d-", offset);

        t->offset = offset;
        soup_message_set_request_header(message, "Range", range);
    }

    if (!offset && gui_prefs_get_cache_enabled()) {
        const char *modtime;

        modtime = gui_cache_get_modification_time(t->cache, t->filename);
        if (modtime) {
            /* We want to get a 304 if we already have the file */
            soup_message_set_request_header(
                message, "If-Modified-Since", modtime);

            /* Handler for 304 Not Modified messages */
            soup_message_add_response_code_handler(
                message, 304, SOUP_HANDLER_PRE_BODY,
                http_response_not_modified, t);
        }
    }

    /* We want to get the chunks out seperately */
    soup_message_set_flags(message, SOUP_MESSAGE_OVERWRITE_CHUNKS);

    /* Handler for normal 200 OK messages */
    soup_message_add_response_code_handler(
        message, 200, SOUP_HANDLER_PRE_BODY, http_response_ok, t);

    /* Handler for 206 Partial Content messages */
    soup_message_add_response_code_handler(
        message, 206, SOUP_HANDLER_PRE_BODY, http_response_partial_content, t);

    soup_message_add_header_handler(
        message, "Content-Length", SOUP_HANDLER_PRE_BODY,
        http_content_length, t);

    soup_message_add_header_handler(
        message, "Content-Range", SOUP_HANDLER_PRE_BODY,
        http_content_range, t);

    soup_message_add_body_handler(
        message, SOUP_HANDLER_PRE_BODY, http_info, t);

    soup_message_add_body_handler(
        message, SOUP_HANDLER_BODY_CHUNK,
        http_read_data, t);

    soup_message_set_request_header(
        message, "User-Agent", "Red Carpet/"VERSION);

    rc_debug(
        RC_DEBUG_LEVEL_INFO, "[%p]: Queuing up new transfer\n",
        message);

    soup_context_unref(context);
    soup_message_queue(message, http_done, t);

    return 0;
} /* http_open */

typedef int (*GUITransferOpenFunc)(GUITransfer *t, int offset);
typedef void (*GUITransferFunc)(GUITransfer *t);

typedef struct {
    const char *identifier;
    gboolean    cacheable;

    GUITransferOpenFunc       open_func;
    GUITransferFunc           abort_func;
    GUITransferFunc           pause_func;
} GUITransferVTable;

GUITransferVTable gui_transfer_vtable[] = {
    { "http",  TRUE,  http_open, http_abort, http_pause },
    { "https", TRUE,  http_open, http_abort, http_pause },
    { "file",  FALSE, file_open, file_abort, file_pause },
    { NULL },
};

static GUITransferVTable *
gui_transfer_get_vtable_from_url(const char *url)
{
    GUITransferVTable *v;

    g_return_val_if_fail(url, NULL);

    v = gui_transfer_vtable;
    while (v->identifier) {
        if (g_strncasecmp(url, v->identifier, strlen(v->identifier)) == 0)
            return v;

        v++;
    }

    g_warning("No idea how to handle URL: %s", url);
    return NULL;
} /* gui_transfer_get_vtable_from_url */

static GUITransferVTable *
gui_transfer_get_vtable(GUITransfer *t)
{
    g_return_val_if_fail(t, NULL);
    g_return_val_if_fail(GUI_IS_TRANSFER(t), NULL);
    g_return_val_if_fail(t->url, NULL);

    return gui_transfer_get_vtable_from_url(t->url);
} /* gui_transfer_get_vtable */

GByteArray *
gui_transfer_begin(GUITransfer *t, const char *url)
{
    GUITransferVTable *vtable;
    int rc;
    
    g_return_val_if_fail(t, NULL);
    g_return_val_if_fail(GUI_IS_TRANSFER(t), NULL);
    g_return_val_if_fail(url, NULL);
    
    vtable = gui_transfer_get_vtable_from_url(url);
    if (!vtable) {
        gui_transfer_set_error(t, GUI_TRANSFER_ERROR_INVALID_URI, url);
        return NULL;
    }

    t->url = g_strdup(url);
    /* Ew. */
    t->filename = g_strdup(strchr(url + strlen(vtable->identifier) + 3, '/'));
    if (!t->filename)
        t->filename = g_strdup(url + strlen(vtable->identifier) + 3);

    /* Reset the offset to make sure nothing icky is going on. */
    t->offset = 0;

    t->cached = FALSE;

    t->cacheable = vtable->cacheable;

    rc_debug(RC_DEBUG_LEVEL_INFO, "Transfer URL: %s\n", t->url);

    rc = (vtable->open_func)(t, 0);
    if (rc) {
        /* An error occurred in the open call. It's the open call's
           responsibility to gui_transfer_set_error() the appropriate
           error. */
        return NULL;
    }

    current_transfers = g_slist_append(current_transfers, t);

    if (t->flags & GUI_TRANSFER_FLAGS_BLOCK) {
        GByteArray *ret;

        gtk_main();

        ret = t->data;
        t->data = NULL;

        return ret;
    }

    return NULL;
} /* gui_transfer_begin */

void
gui_transfer_abort(GUITransfer *t)
{
    GUITransferVTable *vtable;

    g_return_if_fail(t);
    g_return_if_fail(GUI_IS_TRANSFER(t));
   
    vtable = gui_transfer_get_vtable(t);
    g_return_if_fail(vtable);

    t->aborted = TRUE;

    (vtable->abort_func)(t);
} /* gui_transfer_abort */

void
gui_transfer_pause(GUITransfer *t)
{
    GUITransferVTable *vtable;

    g_return_if_fail(t);
    g_return_if_fail(GUI_IS_TRANSFER(t));
    g_return_if_fail(!t->paused);

    vtable = gui_transfer_get_vtable(t);
    g_return_if_fail(vtable);

    if (vtable->pause_func) {
        char *s1, *s2, *str;

        /* Pause it... */
        t->paused = TRUE;
        (vtable->pause_func)(t);

        /* Now do evil formatting */
        if (t->progress_bar) {
            if (t->file_size > 0) {
                s1 = format_size(t->trans_size + t->offset);
                s2 = format_size(t->file_size);
                str = g_strdup_printf(
                    _("%%p%%%% (%s of %s) - Paused"), s1, s2);
                g_free(s1);
                g_free(s2);
                gtk_progress_set_format_string(
                    GTK_PROGRESS(t->progress_bar), str);
                g_free(str);
            }
            else {
                s1 = format_size(t->trans_size + t->offset);
                str = g_strdup_printf(_("%s - Paused"), s1);
                g_free(s1);
                gtk_progress_set_format_string(
                    GTK_PROGRESS(t->progress_bar), str);
                g_free(str);
            }
        }
    }
    else {
        g_warning(
            "Pause not implemented for \"%s\" transfer type",
            vtable->identifier);
    }
} /* gui_transfer_pause */
    
void
gui_transfer_resume(GUITransfer *t)
{
    GUITransferVTable *vtable;

    g_return_if_fail(t);
    g_return_if_fail(GUI_IS_TRANSFER(t));
    g_return_if_fail(t->paused);

    vtable = gui_transfer_get_vtable(t);
    g_return_if_fail(vtable);

    t->paused = FALSE;

    (vtable->open_func)(t, t->offset);
} /* gui_transfer_resume */

gboolean
gui_transfer_error_box(GUITransfer *t, const char *message)
{
    char *error;
    const char *m;
    const char *a;
    gboolean retry;

    if (message)
        m = message;
    else
        m = "";

    /* ugggggggh */
    if (t->error == GUI_TRANSFER_ERROR_CANT_AUTHENTICATE &&
        appinfo->priority && !appinfo->rcd_host_url) {
        a = _("<p>Your account for Red Carpet Premium Services could not "
              "be verified.</p>"
              "<p>Please make sure that your premium services account is "
              "enabled.</p>"
              "<p>To continue using Red Carpet without premium services, "
              "<a href=\"updater:disable_priority\">click here</a>. "
              "You can reenable it later by choosing "
              "Settings->Preferences->General and checking the box "
              "labelled \"Enable RC Premium Services\"</p>");
    }
    else
        a = "";

    error = g_strdup_printf(
        _("<p>Red Carpet was unable to complete a file transfer.</p>"
          "%s"
          "%s"
          "<p>Details:<br>"
          "URL: %s<br>"
          "Error: %s</p>"),
        m,
        a,
        t->url,
        gui_transfer_get_error_string(t));
    retry = rc_gui_error_box(error, RC_ERROR_RETRY);
    g_free(error);

    return retry;
} /* gui_transfer_error_box */

static void
print_headers(gpointer key, gpointer value, gpointer data)
{
    rc_debug(RC_DEBUG_LEVEL_INFO, "   %s: %s\n", (char *) key, (char *) value);
} /* print_headers */

void
rc_dump_transfer_info(SoupMessage *message)
{
    char *url;

    url = soup_uri_to_string(soup_context_get_uri(message->context), FALSE);

    rc_debug(RC_DEBUG_LEVEL_INFO,
             "[%p]: Soup URL %s\n", message, url);
    rc_debug(RC_DEBUG_LEVEL_INFO,
             "[%p]: Method: %s\n", message, message->method);
    rc_debug(RC_DEBUG_LEVEL_INFO,
             "[%p]: HTTP Version: %s\n", message,
             soup_message_get_http_version(message) == SOUP_HTTP_1_0 ? 
             "1.0" : "1.1");
    rc_debug(RC_DEBUG_LEVEL_INFO,
             "[%p]: HTTP Response code: %d\n", 
             message, message->response_code);
    rc_debug(RC_DEBUG_LEVEL_INFO,
             "[%p]: HTTP Response phrase: %s\n",
             message, message->response_phrase ? message->response_phrase :
             "(null)");
    if (message->request_headers) {
        rc_debug(RC_DEBUG_LEVEL_INFO,
                 "[%p]: Request headers:\n", message);
        g_hash_table_foreach(message->request_headers, print_headers, NULL);
    }
    else {
        rc_debug(RC_DEBUG_LEVEL_INFO,
                 "[%p]: No request headers\n", message);
    }

    if (message->response_headers) {
        rc_debug(RC_DEBUG_LEVEL_INFO,
                 "[%p]: Response headers:\n", message);
        g_hash_table_foreach(message->response_headers, print_headers, NULL);
    }
    else {
        rc_debug(RC_DEBUG_LEVEL_INFO,
                 "[%p]: No response headers\n", message);
    }

    g_free(url);
} /* dump_transfer_info */
